/***********************************
 *        Timothy Hannafin         *
 *        hannafit@wpi.edu         *
 * Worcester Polytechnic Institute *
 *               2006              *
 * *********************************/

using System;
using System.Timers;
using System.Threading; 

namespace RobotController
{
	/// <summary>
	/// The AutonoDrive class contains a set of rules based on sensor input and robot status that allow the robot to 
	/// navigate autonomously and avoid obstacles
	/// </summary>
	public class AutonoDriver
	{
		private Robot myRobot;
		private float speed;
		private System.Timers.Timer sensorTimer = new System.Timers.Timer();
		private Script myScript;
		private Script frontScriptR;
		private Script frontScriptL;

		public AutonoDriver(Robot rob, float spd)
		{
			myRobot = rob;
			speed = spd;
		}

		/// <summary>
		/// Start the autonomous driver
		/// </summary>
		public void Start()
		{
			myScript = new Script("script.dat", myRobot, speed);
			sensorTimer.Elapsed += new System.Timers.ElapsedEventHandler(sensor);
			sensorTimer.Interval = 50;
			sensorTimer.Start();
			myRobot.setFowardSpeed(speed);

			//these scripts contain a set of instrctions to execute if something is blocking in front
			frontScriptR = new Script("frontscriptR.dat", myRobot, speed);
			frontScriptL = new Script("frontscriptL.dat", myRobot, speed);


		}

		/// <summary>
		/// This is the main autonomous function it polls the sensors every 50ms and adjusts 
		/// behavior based on the sensor values and current robot status
		/// </summary>
		private void sensor(object source, ElapsedEventArgs e) 
		{
			float front = myRobot.getFrontSensorValue();
			float left = myRobot.getLeftSensorValue();
			float right = myRobot.getRightSensorValue();
			float bottom = myRobot.getBottomSensorValue();

			if(myRobot.Status == StatusType.navigating)
			{
				if(front >=300)  //if something is in front of us turn in the open direction
				{
					sensorTimer.Stop();

					if(left <= right)
					{
						frontScriptL.executeNextBlock();
					}
					else if(right < left)
					{
						frontScriptR.executeNextBlock();
					}

					sensorTimer.Start();
					return;
				}
				if(left > 200)  //if close to a wall on the left increase the left side speed
				{
					myRobot.setLeftSideSpeed(speed+((50-speed)*(left/500)/2));
					return;
				}

				if(right > 200)  //if close to a wall on the right side increase right side speed
				{
					myRobot.setRightSideSpeed(speed+((50-speed)*(right/500)));
					return;
				}

				if(bottom < 100)  //shiny line, execute block command
				{
					myScript.executeNextBlock();
				
				}
				else  //just go straight
				{
					myRobot.setFowardSpeed(speed);
				}
			
			}
			if(myRobot.Status == StatusType.executing_command)
			{

				if(front >=300)  //if something is in front of us turn in the open direction
				{
					myRobot.Status = StatusType.navigation_error;
					myScript.SkipToEndOfBlock();
					this.Stop();
					return;
				}
				if(left > 300)  //if close to a wall on the left increase the left side speed
				{
					myRobot.Status = StatusType.navigation_error;
					myScript.SkipToEndOfBlock();
					this.Stop();
					return;
				}

				if(right > 300)  //if close to a wall on the right side increase right side speed
				{
					myRobot.Status = StatusType.navigation_error;
					myScript.SkipToEndOfBlock();
					this.Stop();
					return;
				}


			}
		}


		/// <summary>
		/// Stop the AutonoDriver
		/// </summary>
		public void Stop()
		{
			sensorTimer.Stop();
		}
	}



}
