/***********************************
 *        Timothy Hannafin         *
 *        hannafit@wpi.edu         *
 * Worcester Polytechnic Institute *
 *               2006              *
 * *********************************/

using System;
using System.Timers;
using System.Threading; 

namespace RobotController
{
	/// <summary>
	/// Robot is the base class for hardware interfacing.  It provides a wrapper for GroupLab.Phidgets calls
	/// </summary>
	/// 		
	public enum StatusType  //status indicator for the robot
	{
		idle, navigating, executing_command, executing_block, anomoly_detected, user_control, navigation_error
	}

	public class Robot
	{


		public StatusType Status = StatusType.idle;
		private GroupLab.Phidgets.InterfaceKit interfaceKit1;
		private GroupLab.Phidgets.UI.InterfaceKitSkin interfaceKitSkin1;
		private GroupLab.Phidgets.Servo servo1;
		private StatusType oldStatus;  //use this to store the old status before executing timed commands

		private System.Timers.Timer commandTimer = new System.Timers.Timer();

		public Robot()
		{
			this.interfaceKit1 = new GroupLab.Phidgets.InterfaceKit();
			this.interfaceKitSkin1 = new GroupLab.Phidgets.UI.InterfaceKitSkin();

			commandTimer.Elapsed += new System.Timers.ElapsedEventHandler(robotCommandTimer);
			commandTimer.Enabled = false;

			this.servo1 = new GroupLab.Phidgets.Servo();
			// 
			// interfaceKit1
			// 
			this.interfaceKit1.AutoAttach = true;
			this.interfaceKit1.PhidgetDevice = null;

			// 
			// servo1
			// 
			this.servo1.AutoAttach = true;
			this.servo1.PhidgetDevice = null;
		}

		public GroupLab.Phidgets.InterfaceKit getInterfaceKit()
		{
			return interfaceKit1;

		}

		#region get sensor value functions

		public float getLeftSensorValue()
		{
			return interfaceKit1.Sensors[0].Value;
		}
		public float getRightSensorValue()
		{
			return interfaceKit1.Sensors[2].Value;
		}
		public float getFrontSensorValue()
		{
			return interfaceKit1.Sensors[1].Value;
		}
		public float getBottomSensorValue()
		{
			return interfaceKit1.Sensors[3].Value;
		}
		#endregion

		#region robot movement commands
		public void stop()
		{
			commandTimer.Enabled = false;
			try
			{
				servo1.Motors[1].Position = servo1.Motors[2].Position = 120;
			}
			catch(Exception e)//System.Runtime.InteropServices.COMException e)
			{
				servo1.Motors[1].Position = 120;
				servo1.Motors[2].Position = 120;
			}
		}

		public void setFowardSpeed(float speed)
		{
			setLeftSideSpeed(speed);
			setRightSideSpeed(speed);
		}
		public void setReverseSpeed(float speed)
		{
			try
			{
				servo1.Motors[1].Position = 135 + speed;
				servo1.Motors[2].Position = 104 - speed;
			}
			catch(System.Runtime.InteropServices.COMException e)
			{	
				servo1.Motors[1].Position = 135 + speed;
				servo1.Motors[2].Position = 104 - speed;
			}

		}
		public void setLeftSideSpeed(float speed)
		{
			try
			{
				servo1.Motors[1].Position = 104 - speed;
			}
			catch(System.Runtime.InteropServices.COMException e)
			{	
				servo1.Motors[1].Position = 104 - speed;
			}
		}

		public void setRightSideSpeed(float speed)
		{
			try
			{
				servo1.Motors[2].Position = 135 + speed;
			}
			catch(System.Runtime.InteropServices.COMException e)
			{	
				servo1.Motors[2].Position = 135 + speed;
			}
		}

		public void turnRight(float speed)
		{
			setLeftSideSpeed(speed);
			setRightSideSpeed(-1*(speed + 31));

		}

		public void turnLeft(float speed)
		{
			setLeftSideSpeed(-1*(speed + 31));
			setRightSideSpeed(speed);

		}



		/***********************************************************
		 * these functions move the robot for a set amount of time
		 * while they are being executed the robot status is changed to 
		 * StatusType.executing_command then changed back to whatever the
		 * status was before
		 * **********************************************************/
		public void setFowardSpeed(float speed, double time)
		{
			oldStatus = this.Status;
			this.Status = StatusType.executing_command;

			setLeftSideSpeed(speed);
			setRightSideSpeed(speed);
			commandTimer.Interval = time;
			commandTimer.Start();
		}

		public void setReverseSpeed(float speed, double time)
		{
			oldStatus = this.Status;
			this.Status = StatusType.executing_command;

			servo1.Motors[1].Position = 135 + speed;
			servo1.Motors[2].Position = 104 - speed;
			commandTimer.Interval = time;
			commandTimer.Start();

		}

		public void turnRight(float speed, double time)
		{
			oldStatus = this.Status;
			this.Status = StatusType.executing_command;

			setLeftSideSpeed(speed);
			setRightSideSpeed(-1*(speed + 31));
			commandTimer.Interval = time;
			commandTimer.Start();
		}

		public void turnLeft(float speed, double time)
		{
			oldStatus = this.Status;
			this.Status = StatusType.executing_command;

			setLeftSideSpeed(-1*(speed + 31));
			setRightSideSpeed(speed);
			commandTimer.Interval = time;
			commandTimer.Start();
		}
		#endregion


		private void robotCommandTimer(object source, ElapsedEventArgs e)  //this timer is used for executing timed commands
		{
			stop();  //stop movement
			this.Status = this.oldStatus;  //change status back to what it was before
			Monitor.PulseAll(this.Status);
			commandTimer.Stop();  //stop the timer
		}
	}
}
